pub static mut LEVEL: Level = Level::Info;

#[allow(dead_code)]
#[repr(u8)]
#[derive(Debug, Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum Level {
    Off = 0,
    Error,
    Warn,
    #[default]
    Info,
    Help,
    Debug
}

const C_RESET: &'static str = "\x1B[0m";
const C_ERROR: &'static str = "\x1B[1;31m";
const C_WARN:  &'static str = "\x1B[1;33m";
const C_INFO:  &'static str = "\x1B[1;32m";
const C_DEBUG: &'static str = "\x1B[1;35m";
const C_HELP:  &'static str = "\x1B[1;36m";

pub fn _log(dbg_loc: Option<(&'static str, u32, u32)>, level: Level, str: &str) {
    if level > unsafe { LEVEL }{
        return;
    }
    let dbg = if let Some((file, line, col)) = dbg_loc {
        format!("{file}:{line}:{col}: ")
    } else {String::new()};
    match level {
       Level::Off => return,
       Level::Error => println!("{dbg}{C_ERROR}error{C_RESET}: {str}"),
       Level::Warn  => println!("{dbg}{C_WARN}warn{C_RESET}: {str}"),
       Level::Info  => println!("{dbg}{C_INFO}info{C_RESET}: {str}"),
       Level::Help  => println!("{dbg}{C_HELP}help{C_RESET}: {str}"),
       Level::Debug => println!("{dbg}{C_DEBUG}debug{C_RESET}: {str}"),
    }
}

pub fn _log_with_loc(dbg_loc: Option<(&'static str, u32, u32)>, loc: String, level: Level, str: &str) {
    if level > unsafe { LEVEL }{
        return;
    }
    let dbg = if let Some((file, line, col)) = dbg_loc {
        format!("{file}:{line}:{col}: ")
    } else {String::new()};
    match level {
        Level::Off => return,
        Level::Error => println!("{dbg}{loc}: {C_ERROR}error{C_RESET}: {str}"),
        Level::Warn  => println!("{dbg}{loc}: {C_WARN}warn{C_RESET}: {str}"),
        Level::Info  => println!("{dbg}{loc}: {C_INFO}info{C_RESET}: {str}"),
        Level::Help  => println!("{dbg}{loc}: {C_HELP}help{C_RESET}: {str}"),
        Level::Debug => println!("{dbg}{loc}: {C_DEBUG}debug{C_RESET}: {str}"),
    }
}

#[macro_use]
pub mod log {
    #[macro_export]
    macro_rules! error {
        ($($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log(Some((file!(), line!(), column!())), crate::logger::Level::Error, &format!($($arg)*))
            } else {
                crate::logger::_log(None, crate::logger::Level::Error, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! warn {
        ($($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log(Some((file!(), line!(), column!())), crate::logger::Level::Warn, &format!($($arg)*))
            } else {
                crate::logger::_log(None, crate::logger::Level::Warn, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! info {
        ($($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log(Some((file!(), line!(), column!())), crate::logger::Level::Info, &format!($($arg)*))
            } else {
                crate::logger::_log(None, crate::logger::Level::Info, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! help {
        ($($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log(Some((file!(), line!(), column!())), crate::logger::Level::Help, &format!($($arg)*))
            } else {
                crate::logger::_log(None, crate::logger::Level::Help, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! debug {
        ($($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log(Some((file!(), line!(), column!())), crate::logger::Level::Debug, &format!($($arg)*))
            } else {
                crate::logger::_log(None, crate::logger::Level::Debug, &format!($($arg)*))
            }
        };
    }


    #[macro_export]
    macro_rules! lerror {
        ($loc:expr, $($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log_with_loc(Some((file!(), line!(), column!())), format!("{}", $loc), crate::logger::Level::Error, &format!($($arg)*))
            } else {
                crate::logger::_log_with_loc(None, format!("{}", $loc), crate::logger::Level::Error, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! lwarn {
        ($loc:expr, $($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log_with_loc(Some((file!(), line!(), column!())), format!("{}", $loc), crate::logger::Level::Warn, &format!($($arg)*))
            } else {
                crate::logger::_log_with_loc(None, format!("{}", $loc), crate::logger::Level::Warn, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! linfo {
        ($loc:expr, $($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log_with_loc(Some((file!(), line!(), column!())), format!("{}", $loc), crate::logger::Level::Info, &format!($($arg)*))
            } else {
                crate::logger::_log_with_loc(None, format!("{}", $loc), crate::logger::Level::Info, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! lhelp {
        ($loc:expr, $($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log_with_loc(Some((file!(), line!(), column!())), format!("{}", $loc), crate::logger::Level::Help, &format!($($arg)*))
            } else {
                crate::logger::_log_with_loc(None, format!("{}", $loc), crate::logger::Level::Help, &format!($($arg)*))
            }
        };
    }
    #[macro_export]
    macro_rules! ldebug {
        ($loc:expr, $($arg:tt)*) => {
            if cfg!(debug_assertions) {
                crate::logger::_log_with_loc(Some((file!(), line!(), column!())), format!("{}", $loc), crate::logger::Level::Debug, &format!($($arg)*))
            } else {
                crate::logger::_log_with_loc(None, format!("{}", $loc), crate::logger::Level::Debug, &format!($($arg)*))
            }
        };
    }
}

