use std::fmt::Display;

use anyhow::bail;

use crate::{common::{loc::LocBox, Loc}, logger::log, validator::predefined::TypeType};

use super::{expr::Expr, literal::Literal, Ident, Program};

#[derive(Debug, Clone)]
pub enum Type {
    Ref { 
        inner: Box<Type>,
        mutable: bool,
    },
    SizedArray {
        inner: Box<Type>,
        count: LocBox<Expr>,
    },
    UnsizedArray {
        inner: Box<Type>,
    },
    Owned(Ident),
}

impl Type {
    pub fn size_of(&self, program: &Program) -> anyhow::Result<usize> {
        match self {
            Self::Ref { inner, mutable } => {
                // TODO: Use the actual ptr size
                Ok(size_of::<*const ()>())
            }
            Self::UnsizedArray { .. } => {
                bail!("Unsized arrays dont have a known size and must be behind a pointer")
            }
            Self::SizedArray { inner, .. } => {
                Ok(inner.size_of(program)? )
            }
            Self::Owned(name) => {
                if let Some(v) = program.structs.get(&name) {
                    return Ok(v.inner().get_size(program)?);
                }
                if let Some(v) = program.types.get(&name) {
                    match v {
                        TypeType::Normal(v) => {
                            return Ok(v.inner().size_of(program)?);
                        }
                        TypeType::Builtin { size, .. } => {
                            return Ok(*size);
                        }
                    }
                }
                if let Some(_) = program.enums.get(&name) {
                    return Ok(4); // TODO: Make enum size changeable
                }
                bail!("Unknown type '{name}'")
            }
        }
    }
    pub fn size_of_allow_unsized_arrays(&self, program: &Program) -> anyhow::Result<usize> {
        match self.size_of(program) {
            Ok(v) => Ok(v),
            Err(e) => {
                if e.to_string().as_str() == "Unsized arrays dont have a known size and must be behind a pointer" {
                    return Ok(0);
                }
                Err(e)
            }
        }
    }
}

impl Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::Ref { inner, mutable } => {
                if *mutable {
                    write!(f, "&mut {inner}")
                } else {
                    write!(f, "&{inner}")
                }
            }
            Self::UnsizedArray { inner } => {
                write!(f, "[{inner}]")
            }
            Self::SizedArray { inner, count } => {
                match count.inner() {
                    Expr::Path(p) => {
                        write!(f, "[{inner}; {p}]")
                    }
                    Expr::Literal(Literal::Number(n)) => {
                        write!(f, "[{inner}; {n}]")
                    }
                    _ => unreachable!()
                }
            }
            Self::Owned(ident) => write!(f, "{ident}")
        }
    }
}
