use std::{fs::File, path::{Path, PathBuf}};
use crate::{cli::CliArgs, parser::ast::Program};

mod x86_64;
mod none;

pub fn get_default_target() -> String {
    x86_64::cgen::linux::CGen::get_target_triple().to_string()
}

pub fn get_all_targets() -> Vec<&'static str> {
    vec![
        // x86_64::asmgen::linux::AsmGen::get_target_triple(),
        // x86_64::cgen::linux::CGen::get_target_triple(),
        // x86_64::llvmgen::linux::LlvmGen::get_target_triple(),
        none::luagen::cct::LuaGen::get_target_triple(),
        x86_64::cgen::linux::CGen::get_target_triple(),
    ]
}

fn get_target_from_triple(triple: &str) -> Box<dyn Target> {
    match triple {
        // _ if triple == x86_64::asmgen::linux::AsmGen::get_target_triple() => {
        //     Box::new(x86_64::asmgen::linux::AsmGen::new())
        // }
        _ if triple == x86_64::cgen::linux::CGen::get_target_triple() => {
            Box::new(x86_64::cgen::linux::CGen::new())
        }
        // _ if triple == x86_64::qbegen::linux::QbeGen::get_target_triple() => {
        //    Box::new(x86_64::qbegen::linux::QbeGen::new())
        // }
        _ => unreachable!()
    }
}

pub fn compile(cli: &CliArgs, programs: Vec<(PathBuf, Program)>) -> anyhow::Result<()> {
    let mut target = get_target_from_triple(&cli.target);
    let build_dir = PathBuf::from("./");
    let mut objs = Vec::new();
    for (fp, program) in programs {
        let int_p = build_dir.join(&fp).with_extension(target.get_int_ext());
        let obj_p = build_dir.join(&fp).with_extension("o");
        
        if !int_p.exists() {
            std::fs::File::create(&int_p)?;
        }

        let mut f = std::fs::File::options()
            .write(true)
            .create(true)
            .truncate(true)
            .open(&int_p)?;
        target.write_code(&program, &mut f)?;
        target.compile(&int_p, &obj_p)?;
        objs.push(obj_p);
    }
    let out = Path::new(&cli.output);
    target.link(objs, &out)?;
    Ok(())
}  

pub trait Target {
    fn new() -> Self where Self: Sized;
    /// Get extension for intermediate file (aka 'asm' or 'c')
    fn get_int_ext(&self) -> &'static str {
        ""
    }
    fn get_target_triple() -> &'static str where Self: Sized {
        ""
    }
    fn write_code(&mut self, program: &Program, f: &mut File) -> anyhow::Result<()>;
    fn compile(&mut self, from: &Path, to: &Path) -> anyhow::Result<()>;
    fn link(&mut self, from: Vec<PathBuf>, to: &Path) -> anyhow::Result<()>;
}
