use std::{fs::File, ops::Deref};
use std::io::Write;
use crate::{common::Loc, parser::ast::{expr::Expr, literal::Literal, statement::{Function, Statement, Struct}, typ::Type, Ast, Ident, Program, Punctuation}, targets::Target, validator::predefined::TypeType};

const INTERNAL_CODE: &'static str = include_str!("internals.c");

pub struct CGen;

impl Target for CGen {
    fn new() -> Self where Self: Sized {
        Self
    }
    fn get_target_triple() -> &'static str {
        "x86_64-cgen-linux"
    }
    fn get_int_ext(&self) -> &'static str {
        "c"
    }
    fn write_code(&mut self, program: &crate::parser::ast::Program, f: &mut File) -> anyhow::Result<()> {
        self.write_fat_comment(f, "Internal number types");
        for (name, typ) in &program.types {
            if let TypeType::Builtin { size: _, signed: _ } = typ {
                self.write_internal_type(f, program, name, typ)?;
            }
        }

        writeln!(f, "{}", INTERNAL_CODE)?;

        self.write_fat_comment(f, "Struct forward declarations");
        for (name, _) in &program.structs {
            writeln!(f, "typedef struct {name} {name};")?;
        }


        self.write_fat_comment(f, "Type aliases");
        for (name, typ) in &program.types {
            if let TypeType::Normal(t) = typ {
                self.write_typedef(f, program, t.inner(), name)?;
            }
        }
        
        self.write_fat_comment(f, "Struct definitions");
        for (name, strct) in &program.structs {
            self.write_struct(f, program, strct.inner(), name)?;
        }

        self.write_fat_comment(f, "Function forward declarations");
        fn write_fn_fwdclr(slf: &mut CGen, program: &Program, f: &mut File, name: &Ident, func: &Function) -> anyhow::Result<()> {
            writeln!(f, "// {}", func.get_def_as_str())?;
            if let Some(rett) = &func.ret_type {
                slf.write_type(f, program, rett.inner())?;
            } else {
                write!(f, "void")?;
            }
            write!(f, " {}(", func.get_full_name())?;
            
            let mut first = true;
            for (name, typ) in &func.params {
                if !first {
                    write!(f, ", ")?;
                }
                slf.write_type(f, program, typ.inner())?;
                write!(f, " {name}")?;
                first = false;
            }
            writeln!(f, ");")?;
            Ok(())
        }
        for (name, func) in &program.functions {
            // skip the main function cause the declaration will be included in the internals
            if name.0 == "main" {
                continue;
            }
            write_fn_fwdclr(self, program, f, name, func.inner())?;
        }
        for (sname, funcs) in &program.member_functions {
            for (fname, func) in funcs {
                write_fn_fwdclr(self, program, f, &Ident(format!("{sname}::{fname}")), func.inner())?;
            }
        }

        

        for i in &program.ast.0 {
            match i {

                Ast::Statement(stat) => {
                    match stat.inner() {
                        Statement::Fn(func) => self.write_fn(f, program, func, stat.loc())?,
                        _ => (),
                    }
                }
                Ast::Expr(_) => unreachable!(),
            }
    
        }
        Ok(())
    }
    fn compile(&mut self, _from: &std::path::Path, _to: &std::path::Path) -> anyhow::Result<()> {
        Ok(())
    }
    fn link(&mut self, _from: Vec<std::path::PathBuf>, _to: &std::path::Path) -> anyhow::Result<()> {
        Ok(())
    }
}

impl CGen {
    fn write_fat_comment(&mut self, f: &mut File, s: &str) {
        let _ = writeln!(f, "\n/**");
        let _ = writeln!(f, " * {s}");
        let _ = writeln!(f, " */\n");
    }
    fn write_expr(&mut self, f: &mut File, program: &Program, expr: &Expr, loc: &Loc) -> anyhow::Result<()> {
        // https://c9x.me/compile/doc/il.html#Memory
        match expr {
            Expr::Literal(lit) => {
                match lit {
                    _ => todo!(),
                }
            }
            Expr::If(_ifexpr) => {
                    
            }
            Expr::BinOp { typ, left, right } => {
                match typ {
                    Punctuation::Eq => {
                        
                    }
                    _ => todo!()
                }
            }
            Expr::Path(p) => {
                for (n, p) in p.0.iter().enumerate() {
                    if n > 0 {
                        write!(f, "::")?;
                    }
                    write!(f, "{p}")?;
                }
            },
            Expr::Return(ret) => {
                write!(f, "return ")?;
                if let Some(ret) = ret.deref() {
                   self.write_expr(f, program, ret.inner(), loc)?;
                }
                writeln!(f, ";")?;
            },
            Expr::Struct { path, fields } => {
                writeln!(f, "({path}) {{")?;
                for (name, val) in fields {
                    write!(f, "    .{name} = ")?;
                    self.write_expr(f, program, val.inner(), val.loc())?;
                    writeln!(f, ",")?;
                }
                write!(f, "}}")?;
            }
            _ => todo!("{expr:?}")
        }
        Ok(())
    }
    
    fn write_typedef(&mut self, f: &mut File, program: &Program, typ: &Type, name: &Ident) -> anyhow::Result<()> {
        fn part(f: &mut File, program: &Program, typ: &Type) -> anyhow::Result<String> {
            let mut pf = String::new();
            match typ {
                Type::Ref { inner, mutable } => {
                    pf.push_str(&part(f, program, inner)?);
                    write!(f, "*")?;
                    if !*mutable {
                        write!(f, "const")?;
                    }
                }
                Type::Owned(name) => {
                    write!(f, "{name}")?;
                }
                Type::SizedArray { inner, count } => {
                    pf.push_str(&part(f, program, inner)?);
                    let Expr::Literal(cnt) = count.inner() else {panic!()};
                    let Literal::Number(cnt) = cnt else {panic!()};
                    pf.push_str(&format!("[{}]", cnt.val));
                }
                Type::UnsizedArray { inner } => {
                    pf.push_str(&part(f, program, inner)?);
                    pf.push_str("[]");
                }
            }
            Ok(pf)
        }

        write!(f, "typedef ")?;
        let pf = part(f, program, typ)?;
        writeln!(f, " {name}{pf};")?;
        
        Ok(()) 

    }
    fn write_type(&mut self, f: &mut File, program: &Program, typ: &Type) -> anyhow::Result<()> {
        static mut WAS_REF: bool = false;
        match typ {
            Type::Ref { inner, mutable } => {
                self.write_type(f, program, inner)?;
                write!(f, " *")?;
                if !*mutable {
                    write!(f, "const")?;
                }
                unsafe { 
                    WAS_REF = true;
                }
            }
            Type::Owned(name) => {
                write!(f, "{name}")?;
            }
            Type::SizedArray { inner, count } => {
                self.write_type(f, program, inner)?;
                let Expr::Literal(cnt) = count.inner() else {panic!()};
                let Literal::Number(cnt) = cnt else {panic!()};
                if !unsafe { WAS_REF } {
                    write!(f, "[{}]", cnt.val)?;
                }
            }
            Type::UnsizedArray { inner } => {
                self.write_type(f, program, inner)?;
                if !unsafe { WAS_REF } {
                    write!(f, "[]")?;
                }
            }
        }
        Ok(()) 
    }

    fn write_internal_type(&mut self, f: &mut File, program: &Program, name: &Ident, typ: &TypeType) -> anyhow::Result<()> {
        match typ {
            TypeType::Normal(typ) => unreachable!(),
            TypeType::Builtin{ size, signed } => {
                if *size == 0 {
                    return Ok(());
                }
                write!(f, "typedef ")?;
                if *signed {
                    write!(f, "signed ")?;
                } else {
                    write!(f, "unsigned ")?;
                }
                match size {
                    9.. => unimplemented!(),
                    0 => (),
                    1 => write!(f, "char ")?,
                    2 => write!(f, "short ")?,
                    4 => write!(f, "int ")?,
                    8 => write!(f, "long long ")?,
                    3 | 5 | 6 | 7  => unreachable!(),
                }
                writeln!(f, "{};", name.0)?;
            }
            _ => (),
        }
        Ok(())
    }

    fn write_struct(&mut self, f: &mut File, program: &Program, strct: &Struct, name: &Ident) -> anyhow::Result<()> {
        writeln!(f, "struct {name} {{")?;
        for (name, typ) in &strct.fields {
            write!(f, "    ")?;
            self.write_type(f, program, typ.inner())?;
            writeln!(f, " {name};")?;
        }
        writeln!(f, "}};")?;
        Ok(())
    }

    fn write_fn(&mut self, f: &mut File, program: &Program, func: &Function, loc: &Loc) -> anyhow::Result<()> {
        writeln!(f, "\n\n// {}", func.get_def_as_str())?;
        if let Some(rett) = &func.ret_type {
            self.write_type(f, program, rett.inner())?;
        } else {
            write!(f, "void")?;
        }
        write!(f, " {}(", func.get_full_name())?;
        
        let mut first = true;
        for (name, typ) in &func.params {
            if !first {
                write!(f, ", ")?;
            }
            self.write_type(f, program, typ.inner())?;
            write!(f, " {name}")?;
            first = false;
        }
        write!(f, ")")?;


        if let Some(body) = &func.body {
            writeln!(f, " {{")?;
            for ast in &body.0 {
                match ast {
                    Ast::Expr(expr) => {
                        self.write_expr(f, program, expr.inner(), expr.loc())?
                    }
                    Ast::Statement(stat) => {
                        //todo!()
                    }
                }
            }
            writeln!(f, "}}")?;
        } else {
            write!(f, ";")?;
        }

        Ok(())
    }
}
