use std::fs::File;
use std::io::Write;
use crate::{common::Loc, parser::ast::{expr::Expr, statement::{Function, Statement}, typ::Type, Ast, Ident, Program, Punctuation}, targets::Target, validator::predefined::TypeType};

pub struct QbeGen {
    var_idx: usize,
}

impl Target for QbeGen {
    fn new() -> Self where Self: Sized {
        Self {
            var_idx: 0
        }
    }
    fn get_target_triple() -> &'static str {
        "x86_64-qbegen-linux"
    }
    fn get_int_ext(&self) -> &'static str {
        "ssa"
    }
    fn write_code(&mut self, program: &Program, f: &mut File) -> anyhow::Result<()> {
        for (name, typ) in &program.types {
            self.write_typetype(f, program, name, typ)?;
        }
        for (name, strct) in &program.structs {
            self.write_type_with_size(f, program, name, strct.inner().get_size(program)?)?;
        }

        for i in &program.ast.0 {
            match i {
                Ast::Statement(stat) => {
                    match stat.inner() {
                        Statement::Fn(func) => self.write_fn(f, program, func, stat.loc())?,
                        _ => (),
                    }
                }
                Ast::Expr(_) => unreachable!(),
            }
    
        }

        Ok(())
    }
    fn compile(&mut self, _from: &std::path::Path, _to: &std::path::Path) -> anyhow::Result<()> {
        Ok(())
    }
    fn link(&mut self, _from: Vec<std::path::PathBuf>, _to: &std::path::Path) -> anyhow::Result<()> {
        Ok(())
    }
}

impl QbeGen {
    fn write_expr(&mut self, f: &mut File, program: &Program, expr: &Expr, loc: &Loc) -> anyhow::Result<()> {
        // https://c9x.me/compile/doc/il.html#Memory
        match expr {
            Expr::If(_ifexpr) => {
                    
            }
            Expr::BinOp { typ, left, right } => {
                match typ {
                    Punctuation::Eq => {
                        
                    }
                    _ => todo!()
                }
            }
            _ => todo!("{expr:?}")
        }
        Ok(())
    }
    fn write_typetype(&mut self, f: &mut File, program: &Program, name: &Ident, typ: &TypeType) -> anyhow::Result<()> {
        match typ {
            TypeType::Normal(typ) => {
                writeln!(f, "type :{:<16} = {{ z {} }}", name.0, typ.inner().size_of_allow_unsized_arrays(program)?)?;
            }
            TypeType::Builtin(size) => {
                writeln!(f, "type :{:<16} = {{ z {size} }}", name.0)?;
            }
            _ => (),
        }
        Ok(())
    }

    fn write_type_with_size(&mut self, f: &mut File, program: &Program, name: &Ident, size: usize) -> anyhow::Result<()> {
        writeln!(f, "type :{:<16} = {{ z {size} }}", name.0)?;
        Ok(())
    }

    fn size_to_qbe_size(&self, size: usize) -> Option<&'static str> {
        match size {
            1 => Some("b"),
            2 => Some("h"),
            4 => Some("w"),
            8 => Some("l"),
            _ => None,
        }
    }
    fn write_fn(&mut self, f: &mut File, program: &Program, func: &Function, loc: &Loc) -> anyhow::Result<()> {
        writeln!(f, "\n\n# {}", func.get_def_as_str())?;
        let mut ret = " ";
        if let Some(r) = &func.ret_type {
            if let Some(s) = self.size_to_qbe_size(r.inner().size_of(program)?) {
                ret = s;
            } else {
                match r.inner() {
                    Type::Owned(name) => {
                        ret = &name.0;
                    }
                    _ => unreachable!(),
                }
            }
        }
        let mut args = String::new();
        for (name, typ) in &func.params {
            if !args.is_empty() {
                args.push_str(", ");
            }
            if let Some(s) = self.size_to_qbe_size(typ.inner().size_of(program)?) {
                args.push_str(s);
            } else {
                match typ.inner() {
                    Type::Owned(name) => {
                        args.push_str(&name.0);
                    }
                    _ => unreachable!(),
                }
            }

            args.push_str(&format!(" %{}", name.0));
        }
        writeln!(f, "function {ret} ${}({args}) {{", func.get_full_name())?;
        writeln!(f, "@start")?;
        if let Some(body) = &func.body {
            for ast in &body.0 {
                match ast {
                    Ast::Expr(expr) => {
                        self.write_expr(f, program, expr.inner(), expr.loc())?
                    }
                    Ast::Statement(stat) => {
                        todo!()
                    }
                }
            }
        }

        writeln!(f, "}}")?;
        Ok(())
    }
}
