use std::collections::HashMap;

use anyhow::{anyhow, bail};

use crate::{common::{loc::LocBox, Loc}, parser::ast::{expr::{Block, Expr}, statement::{Statement, TypeAlias}, typ::Type, Ast, Program}};

pub mod predefined;

struct Scope {
    
}

pub fn validate_code(prog: &mut Program) -> anyhow::Result<()> {
    let Block(items) = prog.ast.clone();
    predefined::load_builtin(prog);
    collect_types_and_constants(prog, &items);
    check_that_types_exist_for_items(prog, &items)?;
    //dbg!(&prog.types);
    //dbg!(&prog.structs);
    //dbg!(&prog.enums);
    //dbg!(&prog.member_functions);
    //dbg!(&prog.functions);
    for item in items {
        match item {
            Ast::Statement(stat) => {
                match stat.inner() {
                    Statement::Fn(func) => {}
                    Statement::Let { .. } => {
                        lerror!(stat.loc(), "Let statements are not allowed outside a function");
                        bail!("")
                    }
                    Statement::ConstVar(var) => {
                    }
                    Statement::StaticVar(var) => {
                    }
                    Statement::Enum(enm) => {}
                    Statement::Struct(strct) => {}
                    Statement::TypeAlias(alias) => {}
                }
            }
            Ast::Expr(_) => unreachable!()
        }
    }


    Ok(())
}

fn check_that_types_exist_for_items(prog: &mut Program, items: &Vec<Ast>) -> anyhow::Result<()> {
    let mut errored = false;
    for item in items {
        match item {
            Ast::Statement(stat) => {
                match stat.inner() {
                    Statement::Fn(func) => {
                        for (name, t) in &func.params {
                            if let Err(_) = validate_type(prog, t) {
                                if let Some(sn) = &func.struct_name {
                                    lerror!(t.loc(), "Type '{}', of argument, '{name}' in function '{}::{}' does not exist", t.inner(), sn, func.name);
                                } else {
                                    lerror!(t.loc(), "Type '{}', of argument, '{name}' in function '{}' does not exist", t.inner(), func.name);
                                }
                                errored = true;
                            }
                        }
                        if let Some(ret_typ) = &func.ret_type {
                            if let Err(_) = validate_type(prog, &ret_typ) {
                                lerror!(ret_typ.loc(), "Return type '{}' of function '{}' does not exist", ret_typ.inner(), func.name);
                                errored = true; 
                            }
                        }
                    },
                    Statement::Let { .. } => (),
                    Statement::ConstVar(var) => {
                        validate_type(prog, &var.typ)?; 
                    }
                    Statement::StaticVar(var) => {
                        validate_type(prog, &var.typ)?; 
                    }
                    Statement::Enum(_) => (),
                    Statement::Struct(strct) => {
                        for (name, t) in &strct.fields {
                            if let Err(_) = validate_type(prog, t) {
                                lerror!(t.loc(), "Type '{}', of field, '{}.{name}' does not exist", t.inner(), strct.name);
                                errored = true; 
                            }
                        }
                    }
                    Statement::TypeAlias(alias) => {
                        if let Err(_) = validate_type(prog, &alias.typ) {
                            lerror!(alias.typ.loc(), "Type '{}' of type alias '{}' does not exist", alias.typ.inner(), alias.name);
                            errored = true; 
                        }
                    }
                }
            }
            Ast::Expr(_) => unreachable!()
        }
    }

    if errored {
        bail!("")
    }
    Ok(())
}

fn validate_type(prog: &mut Program, typ: &LocBox<Type>) -> anyhow::Result<()> {
    fn f(prog: &mut Program, typ: &Type, loc: &Loc) -> anyhow::Result<()> {
        match typ {
            Type::SizedArray { inner, .. } |
            Type::UnsizedArray { inner, .. } |
            Type::Ref { inner, .. } => f(prog, inner, loc),
            Type::Owned(typ) => {
                if prog.enums.get(typ).is_some() ||
                    prog.types.get(typ).is_some() ||
                    prog.structs.get(typ).is_some() {
                    Ok(())
                } else {
                    // lerror!(loc, "Could not find type '{}'", typ.0); 
                    bail!("")
                }
            }
        }
    }
    f(prog, typ.inner(), typ.loc())
}



fn collect_types_and_constants(prog: &mut Program, items: &Vec<Ast>) {
    for item in items {
        match item {
            Ast::Statement(stat) => {
                match stat.inner() {
                    Statement::Fn(func)=> {
                        if let Some(struct_name) = &func.struct_name {
                            if let Some(v) = prog.member_functions.get_mut(&struct_name) {
                                v.insert(func.name.clone(), LocBox::new(stat.loc(), func.clone()));
                            } else {
                                let mut v = HashMap::new();
                                v.insert(func.name.clone(), LocBox::new(stat.loc(), func.clone()));
                                prog.member_functions.insert(struct_name.clone(), v);
                            }
                        } else {   
                            prog.functions.insert(func.name.clone(), LocBox::new(stat.loc(), func.clone()));
                        }
                    } 
                    Statement::Enum(enm) => {
                        prog.enums.insert(enm.name.clone(), LocBox::new(stat.loc(), enm.clone()));
                    }
                    Statement::Struct(strct) => {
                        prog.structs.insert(strct.name.clone(), LocBox::new(stat.loc(), strct.clone()));
                    }
                    Statement::TypeAlias(alias) => {
                        let typ = alias.clone().typ.inner().clone();
                        prog.types.insert(alias.name.clone(), predefined::TypeType::Normal(LocBox::new(stat.loc(), typ)));
                    }
                    Statement::Let { .. } => (),
                    Statement::ConstVar(var)  => {
                        prog.const_vars.insert(var.name.clone(), var.clone());
                    }
                    Statement::StaticVar(var) => {
                        prog.static_vars.insert(var.name.clone(), var.clone());
                    },
                }
            }
            Ast::Expr(_) => unreachable!()
        }
    }

}
