use std::collections::HashMap;
use lazy_static::lazy_static;
use crate::common::Loc;
use crate::parser::typ::parse_type;
use crate::{common::loc::LocBox, parser::ast::{statement::Function, typ::Type, Ident, Program}};

#[cfg(target_arch="x86_64")]
const SIZE: usize = 8;
#[cfg(target_arch="x86")]
const SIZE: usize = 4;

lazy_static!(
    pub static ref TYPES_RAW: HashMap<&'static str, (usize, bool)> = [
        ("void",  (0, false)),
        ("usize", (SIZE, false)),
        ("isize", (SIZE, true)),
        ("u8",    (1, false)),
        ("u16",   (2, false)),
        ("u32",   (4, false)),
        ("u64",   (8, false)),
        ("i8",    (1, true)),
        ("i16",   (2, true)),
        ("i32",   (4, true)),
        ("i64",   (8, true)),
    ].into();
    pub static ref FUNCTIONS: HashMap<&'static str, (Vec<(&'static str, &'static str)>, &'static str)> = [
        ("syscall", (vec![
            ("arg_count", "u8"),
            ("sc_num", "u8"),
            ("args", "&[&void]")
        ], "usize")),
    ].into();
);

#[derive(Debug, Clone)]
pub enum TypeType {
    Normal(LocBox<Type>),
    Builtin { 
        size: usize,
        signed: bool,
    }
}

pub fn load_builtin(prog: &mut Program) {
    for (name, (size, signed)) in TYPES_RAW.iter() {
        prog.types.insert(Ident(name.to_string()), TypeType::Builtin{ size: *size, signed: *signed });
    }
    
    for (name, (args, ret_typ)) in FUNCTIONS.iter() {
        let mut params = Vec::new();
        let mut ret_type = None;
        if ret_typ.len() > 0 {
            let mut ret_t_tokens = crate::tokeniser::tokenise(&ret_typ, "(internal)").unwrap();
            let typ = parse_type(&mut ret_t_tokens).unwrap();
            ret_type = Some(LocBox::new(&Loc::default(), typ.inner().clone()));
        }
        for (name, typ) in args {
            let mut tokens = crate::tokeniser::tokenise(&typ, "(internal)").unwrap();
            let typ = parse_type(&mut tokens).unwrap();
            params.push((Ident(name.to_string()), LocBox::new(&Loc::new("(internal)", 0, 0), typ.inner().clone()))); 
        }

        let f = Function {
            struct_name: None,
            name: Ident(name.to_string()),
            params,
            ret_type,
            qual_const: false,
            qual_extern: None,
            body: None
        };
        prog.functions.insert(Ident(name.to_string()), LocBox::new(&Loc::new("(internal)", 0, 0), f));
    }
}
